#!/usr/bin/env python2
# -*- coding: UTF-8 -*-
# --------------------------------------------------------------------------
# Get YouTube Video, Copyright (C) 2006-2011 Otto Allmendinger, Hirs, Mathias Panzenböck
# --------------------------------------------------------------------------
# ***** BEGIN GPL LICENSE BLOCK *****
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
#
# ***** END GPL LICENCE BLOCK *****
# --------------------------------------------------------------------------

# version: 1.12a
#
# differnet versions:
#  by Otto Allmendinger <otto.allmendinger@gmail.com> (original author):
#     http://www.kde-apps.org/content/show.php?content=37509
#  by Hirs:
#     http://www.kde-apps.org/content/show.php?content=41432
#  by panzi (this version):
#     http://www.kde-apps.org/content/show.php?content=41456
#
# further credits:
# spanish translation of the menu entries by Víctor Fernández Martínez
# improvements to video encoding by Tejas Dinkar <tejasdinkar AT gmail DOT com>
# tipps about warnings and errors in python 2.3 by Volker Janzen
# (However, I dropped python 2.3 support now.)
# Russian and Ukrainian translations by Ivan Lezhnev Jr.

import urllib, re, os, sys
import gettext, locale
from subprocess import Popen, PIPE
from htmlentitydefs import name2codepoint
from lxml import html

try:
	from urlparse import parse_qs
except ImportError:
	from cgi import parse_qs

try:
	import json
except ImportError:
	import simplejson as json

import traceback
from ConfigParser import ConfigParser
from getopt import gnu_getopt as getopt

if os.getenv('KDE_SESSION_VERSION') == '4':
	FILE_CLIENT='kioclient'
	KCOOKIEJAR='kcookiejar4'
	try:
		import dbus
	except ImportError:
		GET_COOKIES=lambda: cmd('qdbus', 'org.kde.kded', '/modules/kcookiejar',
			'org.kde.KCookieServer.findDOMCookies', 'http://www.youtube.com')
	else:
		_get_cookies = dbus.SessionBus().get_object('org.kde.kded',
			'/modules/kcookiejar').get_dbus_method('findDOMCookies',
				dbus_interface='org.kde.KCookieServer')
		GET_COOKIES=lambda: (_get_cookies('http://www.youtube.com', 0), '', 0)

else:
	FILE_CLIENT='kfmclient'
	KCOOKIEJAR='kcookiejar'
	GET_COOKIES=lambda: cmd('dcop', 'kcookiejar', 'kcookiejar',
		'findCookies(QString)', 'http://www.youtube.com')

def shesc(*s):
	if len(s) == 1:
		return "'%s'" % s[0].replace("'","'\\''")
	else:
		return tuple(("'%s'" % arg.replace("'","'\\''")) for arg in s)

def strconv(f):
	def __strconv(*args,**kwargs):
		return ''.join(f(*args,**kwargs))
	
	return __strconv

@strconv
def htmlesc(s):
	for c in s:
		if c == '<':
			yield "&lt;"
		elif c == '>':
			yield "&gt;"
		elif c == '"':
			yield "&quot;"
		elif c == '&':
			yield "&amp;"
		else:
			yield c

def htmlmap(s):
	if s[0:1] == '#':
		try:
			if s[1:2] == 'x':
				return unichr(int(s[2:],16))
			else:
				return unichr(int(s[1:],10))
		except ValueError:
			pass

	cp = name2codepoint.get(s,None)

	if cp is None:
		return None

	return unichr(cp)

@strconv
def htmlunesc(s):
	i = 0

	while True:
		j = s.find('&',i)

		if j < 0:
			yield s[i:]
			break

		else:
			yield s[i:j]

			i = j
			j = s.find(';',i)

			if j < 0:
				yield '&'
				i += 1

			else:
				tr = htmlmap(s[i+1:j])

				if tr is None:
					yield '&'
					i += 1

				else:
					yield tr
					i = j + 1

def fixendl(s):
	if s[-1:] == '\n':
		return s[:-1]
	else:
		return s

def cmd(*args,**kwargs):
	p = Popen([unicode(arg).encode(sys.stdin.encoding) for arg in args],**kwargs)

	if kwargs.get('stderr',None) == PIPE:
		serr = fixendl(unicode(p.stderr.read(),sys.stderr.encoding))
	else:
		serr = u''

	if kwargs.get('stdout',None) == PIPE:
		sout = fixendl(unicode(p.stdout.read(),sys.stdout.encoding))
	else:
		sout = u''

	return sout, serr, p.wait()

# dialoges:
def showError(caption,msg):
	return cmd('kdialog', '--caption', caption, '--error', msg)

def showSaveFile(caption,file_name):
	return cmd('kdialog', '--caption', caption, '--getsavefilename', file_name, stdout=PIPE)
	
def showWarningYesNoCancel(caption,msg):
	return cmd('kdialog', '--caption', caption, '--warningyesnocancel', msg)

def showPassivePopup(caption,msg,secs=5):
	return cmd('kdialog', '--caption', caption, '--passivepopup', msg, secs)

def showMenu(caption,msg,options):
	args = ['kdialog', '--caption', caption, '--menu', msg]
	for key, value in options:
		args.append(json.dumps(key))
		args.append(value)
	sout, serr, status = cmd(*args,stdout=PIPE)
	if status == 0:
		return json.loads(sout)
	else:
		return None

def parse_fmt(fmt):
	fmt = fmt.split("/")
	size = fmt[1].split("x")
	return (fmt[0], (int(size[0],10), int(size[1],10)))

def get_video_info(video_id):
	url = "http://www.youtube.com/watch?v=" + video_id
	stream = urlopener.open(url)

	try:
		data = stream.read()
	finally:
		stream.close()
	
	data = unicode(data,"utf-8")
	doc = html.document_fromstring(data)
#	token   = None
	videoid = None
	name    = None
	user    = None
	fmt_list = None
	fmt_stream_map = None
	
	for script in doc.xpath("//script"):
		if not script.text:
			continue
		m = re.search(r"^\s*yt\.playerConfig\s*=\s*(.*);\s*$",script.text,re.M)
		if m:
			try:
				config = json.loads(m.group(1))
			except ValueError:
				pass
			else:
				args = config.get('args',None)
				if args:
					fmt_list = args.get('fmt_list','')
					if not fmt_list:
						args.get('fmt_map','')
					videoid = args.get('video_id',None)
					fmt_stream_map = args.get('url_encoded_fmt_stream_map',None)

	if not fmt_stream_map or not videoid:
		raise ValueError(
			_(u'<p>No video information found on the video\'s Page.<br/>\n'
			   'Maybe you have to be logged in to view this video or YouTube has changed it\'s page layout?<br/>\n'
			   'If you are already logged in or you don\'t have to be logged in to view this video, '
			   'please report this <a href="http://kde-apps.org/content/show.php?content=41456">here</a>.\n'
			   'video-page: <a href="%(url)s">%(url)s</a></p>') %
			{'url':url})

	if not fmt_list:
		fmt_list = {}
	else:
		fmt_list = dict(parse_fmt(fmt) for fmt in fmt_list.split(","))
	fmt_stream_map = [dict((item[0], ''.join(item[1])) for item in parse_qs(info).items()) for info in fmt_stream_map.split(",")]

	video_object = doc.xpath('//*[@itemscope][@itemtype="http://schema.org/VideoObject"]')
	if video_object:
		video_object = video_object[0]
		name_object = video_object.xpath('//meta[@itemprop="name"]/@content')
		if name_object:
			name = name_object[0]
		users = video_object.xpath("//a[@rel='author']")
		if users:
			user = users[0].get("title") or users[0].text or None
	
	return {"id": videoid, "title": name, "user": user, "fmt_stream_map": fmt_stream_map, "fmt_list": fmt_list}

def get_filename(save_dir, video_id, select):
	video_info = get_video_info(video_id)

	videoid = video_info["id"]
	name    = video_info["title"]
	user    = video_info["user"]
	fmt_stream_map = video_info["fmt_stream_map"]
	
	if not name:
		name = videoid
	else:
		name = name
	
	if user:
		name += u" by " + user

	if 'format' in select:
		fmt = select['format']

		if fmt.lower() == "ask":
			menu = []
			fmt_list = video_info["fmt_list"]
			for fmt_info in video_info["fmt_stream_map"]:
				size = fmt_list.get(fmt_info["itag"],None)
				if size:
					fmt_info["size"] = "%dp" % size[1]
				else:
					fmt_info["size"] = fmt_info["quality"]
				fmt_info["format"] = re.sub("^x-", "", fmt_info["type"].split(";")[0].split("/")[1])
				menu.append((fmt_info["itag"],"%(size)s %(format)s" % fmt_info))

			itag = showMenu(_(u"Save YouTube Video"),_(u"Chose format of »%(title)s« by »%(user)s«:") % video_info,menu)
			if itag is None:
				return None

		else:
			try:
				itag = FORMATS[fmt][0]
			except KeyError:
				if fmt.isdigit():
					itag = fmt
				else:
					raise ValueError(_(u'Unknown file format: %(format)s') % {'format': fmt})

		del select['format']
		select['itag'] = itag

	fmt_info = None
	for item in fmt_stream_map:
		sliced = dict((key,item[key]) for key in item if key in select)
		if 'type' in sliced:
			sliced['type'] = sliced['type'].split(";")[0]
		if sliced == select:
			fmt_info = item

	if fmt_info is None:
		raise ValueError(
			_(u'<p>Video "%(name)s" does not support given options:</p><ul>%(options)s</ul>') %
			{"name":name, "options": "\n".join("<li><b>%s:</b> %s</li>" % (key, select[key]) for key in sorted(select))})

	# kdialog seems to have problems with "
	name = re.sub('[/:]', '', name).replace('"', "'")

	return u"%s/%s.%s" % (save_dir,name,EXT_MAP.get(fmt_info["type"].split(";")[0],"flv")), fmt_info['url']

def get_video(vid_url, file_name, action):
	if cmd(FILE_CLIENT, 'copy', vid_url, file_name)[2]:
		raise Exception(_(u'Error saving file: %(file)s') % {'file':file_name})
	
	if action:
		fdir, fname = os.path.split(file_name)
		fname, fext = os.path.splitext(fname)
		cmd_vars = {
			'fdir':  shesc(fdir),
			'fname': shesc(fname),
			'fext':  shesc(fext),
			'fpath': shesc(file_name)
		}
		command = action % cmd_vars

		p = Popen(command,shell=True,stderr=PIPE)
		code = p.wait()

		if code:
			raise Exception(
				_(u"<p>Error executing command:<br/>\n"
				   "<pre>%(command)s</pre><br/>\n"
				   "Command output:<br/>\n"
				   "<pre>%(output)s</pre></p>") %
				{'command':htmlesc(command),'output':htmlesc(p.stderr.read())})

def usage():
	print("""\
Usage: %s [OPTION...] [VIDEO-ID or URL...]

OPTIONS:
	-h, --help             show this help message

    -l, --list             list supported formats of video
	-a, --action=ACTION    execute ACTION after download
	        following referneces may be included in ACTION:
	            %%(fdir)s .... directory to which the file was saved
	            %%(fname)s ... name of the saved file
	            %%(fext)s .... extension of the saved file
	            %%(fpath)s ... full path of the saved file

	-f, --format=FORMAT    file format to download
	        following formats are supported:""" % sys.argv[0])

	for formats, value in _FORMATS:
		print("""\
	            %s:
	                %s
""" % (', '.join(formats), value[1]))

	print("""\
	            <digits>:
	                specify format code directly (unknown)

	            ask:
	                present a menu with available formats

	-i, --itag=ITAG        number
	-t, --type=TYPE        video/x-flv, video/mp4, video/webm, ...
	-q, --quality=QUALITY  small, medium, large, hd720, hd1080, ...
""")

EXT_MAP = {
	"video/x-flv": "flv",
	"video/mp4":   "mp4",
	"video/webm":  "webm"
}

def main():
	global urlopener, _, _FORMATS, FORMATS

	try:
		conf_file = os.path.join(os.getenv('HOME'),'.get_yt_videorc')

		bin_dir    = os.path.dirname(sys.argv[0])
		prefix     = os.path.dirname(bin_dir)
		locale_dir = os.path.join(prefix,'share','get_yt_video')

		LANG = os.getenv('LANG')
		
		if locale.getlocale() == (None, None) and LANG is not None:
			try:
				lang = gettext.translation('get_yt_video', locale_dir, languages=[LANG])
				lang.install()
			except:
				_ = lambda s: s

		else:
			gettext.install('get_yt_video', locale_dir)

		_FORMATS = [
			(['flv'],                   ('5',  _(u'Normal 240p (flv)'))),
			(['flv-hq-'],               ('6',  _(u'Old High Quality 480p (flv)'))),
			(['3gp'],                   ('13', _(u'Mobil Version (3gp)'))),
			(['3gp-hq', '3gp-144p'],    ('17', _(u'Mobil Version 144p (3gp)'))),
			(['mp4-hq', 'mp4-480p'],    ('18', _(u'High Quality 480p (mp4)'))),
			(['mp4-hd', 'mp4-720p'],    ('22', _(u'High Definition 720p (mp4)'))),
			(['mp4-360p'],              ('34', _(u'High Quality 360p (mp4)'))),
			(['flv-hq', 'flv-480p'],    ('35', _(u'High Quality 480p (flv)'))),
			(['mp4-hd+', 'mp4-1080p'],  ('37', _(u'High Definition 1080p (mp4)'))),
			(['mp4-hd++', 'mp4-3072p'], ('38', _(u'High Definition 3072p (mp4)'))),
			(['webm', 'webm-360p'],     ('43', _(u'Normal 360p (webm)'))),
			(['webm-720p'],             ('45', _(u'High Definition 720p (webm)'))),
		]
		
		FORMATS = {}
		
		for keys, value in _FORMATS:
			keys.append(value[1])
			for key in keys:
				FORMATS[key] = value

		# parse command line options:
		try:
			opts, args = getopt(sys.argv[1:], "ha:f:i:t:q:l",
				["help", "action=", "format=", "itag=", "type=", "quality=", "list"])
		except Exception, e:
			sys.stderr.write('*** %s\n' % e)
			sys.exit(2)
		
		action = ''
		select = {}
		list_fmts = False

		for o, a in opts:
			if o in ("-a", "--action"):
				action = a
			elif o in ("-f", "--format"):
				select['format'] = a.strip().lower()
			elif o in ("-i", "--itag"):
				select["itag"] = a.strip().lower()
			elif o in ("-t", "--type"):
				select["type"] = a.strip().lower()
			elif o in ("-q", "--quality"):
				select["quality"] = a.strip().lower()
			elif o in ("-h", "--help"):
				usage()
				return
			elif o in ("-l", "--list"):
				list_fmts = True
			else:
				assert False, "unhandled option"

		if not select:
			select['format'] = 'mp4-hq'
		
		if len(args) < 1:
			sys.stderr.write('*** no url given\n')
			usage()
			sys.exit(2)

		# use a FancyURLopener so we can use konquerors cookies :)
		urlopener = urllib.FancyURLopener()
		
		cookies, serr, code = GET_COOKIES()

		if code != 0:
			# try to start kcookiejar
			print('could not get cookies, trying to start cookie service')
			ignore, serr, code = cmd(KCOOKIEJAR)

			if code == 0:
				cookies, serr, code = GET_COOKIES()
				if code != 0 or not cookies:
					if serr:
						print(serr)
					print('could not get cookies (maybe no cookies set yet)')
			else:
				if serr:
					print(serr)
				print('failed to start cookie service')

		if code == 0 and cookies:
			urlopener.addheader("Cookie",cookies)

		conf = ConfigParser()

		video_ids = []
		
		for url in args:
			# searching for video id in url:
			m = re.search(r'youtube\.com/v/(?P<video_id>[-A-Za-z0-9_]+)', url)
			
			if m is None:
				m = re.search('video_id=(?P<video_id>[-A-Za-z0-9_]+)', url)
				
			if m is None:
				m = re.search('v=(?P<video_id>[-A-Za-z0-9_]+)', url)
			
			if m is not None:
				# video id found
				video_id = m.group('video_id')
				if video_id not in video_ids:
					video_ids.append(video_id)
	
			else:
				# searching for video ids in content:
				stream = urlopener.open(url)
				
				try:
					data = stream.read()
				finally:
					stream.close()
				
				found = re.findall('youtube\.com/v/([-A-Za-z0-9_]+)', data)
				
				if not found:
					found = re.findall('video_id=([-A-Za-z0-9_]+)', data)
				
				if not found:
					found = re.findall('v=([-A-Za-z0-9_]+)', data)
			
				if not found:
					showError(
						_(u"Save YouTube Video"),
						_(u"Can\'t get the youtube-videos ID out of the url or the referred site content.\n"
						   "Maybe you have to be logged in to view this video?"))

				for video_id in found:
					if video_id not in video_ids:
						video_ids.append(video_id)
			
		conf.read([conf_file])

		if conf.has_option('get_yt_video','save_dir'):
			save_dir = conf.get('get_yt_video','save_dir')
		
		else:
			save_dir = os.curdir
		
		if not conf.has_section('get_yt_video'):
			conf.add_section('get_yt_video')

		if list_fmts:
			for video_id in video_ids:
				video_info = get_video_info(video_id)
				print("%(title)s by %(user)s http://www.youtube.com/watch?v=%(id)s" % video_info)
				fmt_list = video_info["fmt_list"]
				for fmt_info in video_info["fmt_stream_map"]:
					size = fmt_list.get(fmt_info["itag"],None)
					if size:
						fmt_info["size"] = "%dx%d" % size
					else:
						fmt_info["size"] = "rsolution unknown"
					print("itag: %(itag)s, quality: %(quality)s (%(size)s), type: %(type)s" % fmt_info)
				print("")
		else:
			for video_id in video_ids:
				file_info = get_filename(save_dir, video_id, select)
				if file_info is None:
					continue

				file_name, vid_url = file_info
				
				while True:
					file_name, serr, code = showSaveFile(_(u"Save YouTube Video"),file_name)

					if code != 0: # Cancel
						break

					if not os.path.exists(file_name):
						break

					ignore, serr, code = showWarningYesNoCancel(
							_(u"Save YouTube Video"),
							_(u'File "%(file)s" already exists.\nDo you want to replace it?') %
							{'file':file_name})
						
					if code == 0: # Yes
						break

					elif code == 1: # No
						continue

					else: # code == 2: Cancel
						break

				if code == 0: # Ok [+ Replace]
					save_dir = os.path.split(file_name)[0]

					try:
						conf.set('get_yt_video','save_dir',save_dir)
						conf.write(open(conf_file,'w'))
					except:
						traceback.print_exc(file=sys.stderr)
			
					try:
						get_video(vid_url, file_name, action)
				
					except Exception, errmsg:
						traceback.print_exc(file=sys.stderr)
						showError(_(u'Save YouTube Video'),str(errmsg))

	except Exception, err:
# debug output:
#		try:
#			from cStringIO import StringIO
#		except ImportError:
#			from StringIO import StringIO
#		sio = StringIO()
#		traceback.print_exc(file=sio)
#		showError(_(u'Save YouTube Video'),"<pre>%s</pre>" % htmlesc(sio.getvalue()))
#		sio.close()
		
		traceback.print_exc(file=sys.stderr)
		showError(_(u'Save YouTube Video'),str(err))

if __name__ == "__main__":
	main()
